
/* audio_examp.q: some simple examples for the audio and sndfile modules */

import audio, sndfile;

/* Audio sample rate, format, number of channels, buffer size and number of
   buffers for audio input and output. We assume a device capable of recording
   and playing back 16 bit integer samples at 44100 Hz in stereo, using 2048
   frame buffers. If you experience drop-outs (e.g., crackling noise in
   recorded wave, or during playback) then try to increase the buffer size or
   experiment with different NBUFS values (the default value of 0 causes
   PortAudio to choose a system-dependent minimum value to prevent audio
   glitches). Under Linux, we also recommend running a low-latency kernel; in
   this case you might achieve lower latencies by setting the
   PA_MIN_LATENCY_MSEC environment variable (see below). */

def RATE = 44100, FORMAT = PA_INT16, CHANNELS = 2, BUFSIZE = 2048, NBUFS = 0,
  INFO = (RATE,CHANNELS,FORMAT,BUFSIZE,NBUFS);

/* Uncomment this to experiment with minimum latency values. This value is
   used by PortAudio to determine the minimum buffer number when NBUFS is
   zero. */

// def _ = setenv "PA_MIN_LATENCY_MSEC" "10"; // minimum latency in ms

/* Soundfile format. This is used when creating new soundfiles with sf_record
   or sf_save. */

def SF_FORMAT = SF_FORMAT_WAV or SF_FORMAT_PCM_16;

/* Check system for available realtime thread priorities. */

def POL = SCHED_FIFO, PRIO = maxprio;

testprio PRIO		= setsched this_thread 0 0 || true
			    where () = setsched this_thread POL PRIO;
			= false otherwise;

maxprio			= until (neg testprio) (+1) 1 - 1;

/* Execute a task in realtime. */

special realtime X;

realtime X		= setsched TH 0 0 || Y
			    where TH = this_thread,
			      Y = setsched TH POL PRIO || X;

/* Record a wave and return it as a byte string. The given TIME parameter can
   be an integer or floating point value denoting the time in seconds to
   record. NOTE: The entire wave is stored in RAM, so you better keep it
   small. At 44100 KHz/16 bit/stereo each minute needs some 10 MB of RAM! */

record TIME		= realtime (read_audio_stream IN LEN)
			    where IN:AudioStream =
			      open_audio_stream AUDIO_IN PA_READ INFO,
			      LEN = round (TIME*RATE*1.0);

/* Play back a wave encoded as a byte string. */

play WAV		= realtime (write_audio_stream OUT WAV)
			    where OUT:AudioStream =
			      open_audio_stream AUDIO_OUT PA_WRITE INFO;

/* Loop a wave. This just keeps playing the same wave over and over again,
   until you hit Ctl-C to stop it. */

loop WAV		= realtime (play_loop OUT WAV)
			    where OUT:AudioStream =
			      open_audio_stream AUDIO_OUT PA_WRITE INFO;

play_loop OUT WAV	= write_audio_stream OUT WAV || play_loop OUT WAV;

/* Safe/load a wave to/from a file. */

sf_save NAME WAV	= sf_write SF WAV
			    where SF:SndFile = sf_open NAME SFM_WRITE
			      (RATE,CHANNELS,SF_FORMAT);

sf_write		= sf_write_float if FORMAT = PA_FLOAT32;
			= sf_write_int if FORMAT = PA_INT32;
			= sf_write_short if FORMAT = PA_INT16;

sf_load NAME		= sf_read SF (sf_frames SF)
			    where SF:SndFile = sf_open NAME SFM_READ ();

sf_read			= sf_read_float if FORMAT = PA_FLOAT32;
			= sf_read_int if FORMAT = PA_INT32;
			= sf_read_short if FORMAT = PA_INT16;

/* Record and play back waves to/from a file. This lets you record and play
   waves which do not fit into RAM. */

sf_record NAME TIME	= realtime (sf_record_loop sf_write SF IN T0 T)
			    where IN:AudioStream =
			      open_audio_stream AUDIO_IN PA_READ INFO,
			      SF:SndFile = sf_open NAME SFM_WRITE
			      (RATE,CHANNELS,SF_FORMAT),
			      T = round (TIME*RATE*1.0),
			      T0 = audio_stream_time IN;

sf_record_loop F SF IN T0 T
			= F SF (read_audio_stream IN BUFSIZE) ||
			  sf_record_loop F SF IN T0 T
			    if audio_stream_time IN-T0 <= T;
			= () otherwise;

sf_play NAME		= realtime (sf_play_loop sf_read SF OUT)
			    where OUT:AudioStream =
			      open_audio_stream AUDIO_OUT PA_WRITE INFO,
			      SF:SndFile = sf_open NAME SFM_READ ();

sf_play_loop F SF OUT	= write_audio_stream OUT WAV || sf_play_loop F SF OUT
			    if #WAV > 0
			    where WAV = F SF BUFSIZE;
			= () otherwise;

/* Encode a wave, given as a list of (tuples of) floating point sample values
   in the range -1..1, as a byte string in the target format. */

encode Xs		= bcat (map encode_frame Xs);

encode_frame Xs:Tuple	= bcat (map encode_sample (list Xs));
encode_frame X		= encode_sample X otherwise; // mono

encode_sample X:Float	= bytestr (X,4)
			    if FORMAT = PA_FLOAT32;
			= bytestr (round (0x7fffffff*X),4)
			    if FORMAT = PA_INT32;
			= bytestr (round (0x7fff*X),2)
			    if FORMAT = PA_INT16;

/* Example: Generate a sine wave with given amplitude, frequency and length.
   E.g., you can encode and loop one cycle of a 440 Hz sine wave as follows:
   loop (encode (sine 0.2 440 (1/440))). */

/* NOTE: For large waves, this is fairly slow, as first enumerating all
   samples in a list and then encoding the samples and concatenating the
   results to a byte string takes its time. Thus generating a lot of samples,
   maybe even in realtime, should be left to specialized external functions
   written in C (until computers become much faster than they are now). */

def PI = 4.0*atan 1.0;

sine AMP FREQ TIME	= map (sine_frame AMP FREQ) (nums 0 (LEN-1))
			    where LEN = round (TIME*RATE*1.0);

sine_frame AMP FREQ X	= tuple (map (cst (sine_sample AMP FREQ X))
				 (nums 1 CHANNELS))
			    if CHANNELS <> 1;
			= sine_sample AMP FREQ X otherwise; // mono

sine_sample AMP FREQ X	= AMP*sin (2*PI*FREQ/RATE*X);

/* Decode a wave. Converts a byte string back to a list of (tuples of) sample
   values in floating point format. */

def BPS = bps FORMAT, BPF = CHANNELS*BPS;

bps FORMAT		= 2 if FORMAT = PA_FLOAT32;
			= 4 if FORMAT = PA_INT32;
			= 2 if FORMAT = PA_INT16;

decode B		= map decode_frame (split_bytestr BPF B);

decode_frame B		= decode_sample B if #B = BPS; // mono
			= tuple (map decode_sample (split_bytestr BPS B));

decode_sample B		= bfloat B if FORMAT = PA_FLOAT32;
			= signed (bint B) / 0x7fffffff if FORMAT = PA_INT32;
			= signed (bint B) / 0x7fff if FORMAT = PA_INT16;

split_bytestr BYTES B	= map (slice B BYTES) (numsby BYTES 0 (#B-1));

slice B BYTES I		= sub B I (I+BYTES-1);

def MSKS = [0x0, 0x7f, 0x7fff, 0x7fffff, 0x7fffffff], MSK = not (MSKS!BPS);

signed N		= N or MSK if N and MSK > 0;
			= N otherwise;
